/*
 *  comgt version 0.xx - 3G/GPRS datacard management utility
 *
 *  Copyright (C) 2003     Paul Hardwick <paul@peck.org>
 *  Copyright (C) 2020-21  Jonas Bechtel <code@jbechtel.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 */

 /***************************************************************************
* $Id: comgt.c,v 1.4 2006/10/20 14:30:19 pharscape Exp $
 ****************************************************************************/

#ifndef COMGT_MAIN
#define COMGT_MAIN

#include <stdbool.h>

#ifndef NOCODE

#define _POSIX_C_SOURCE 199506
#define _DEFAULT_SOURCE
#define _BSD_SOURCE // _BSD_SOURCE is deprecated but only if _DEFAULT_SOURCE isn't set??
#include <unistd.h>
#include <stdio.h>
#include "stdlib-compat.h"
#include <signal.h>
#include <termios.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>
#include <stdint.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <getopt.h>


#include "comgt.h"   // This is just compiled-in script texts
#include "version.h"
const char comgt_version[] = COMGT_VERSION_STRING;

#endif // ifndef NOCODE


#include "bstring-prettyname.h" // bstring2021

struct comgt_dev {
  bstring filename; /* Comm device.  May be "-" */
  int fd;
  bool high_speed; // This makes LOWER speed if set to TRUE
  struct termios work_termio, orig_termio;
};

#ifndef NOCODE

typedef struct comgt_dev* cdev;


#define MAXLABELS 3000  /* Maximum number of labels */
#define MAXGOSUBS 128   /* Max depth */
#define RPIPE_OUTPUT_MAXL 2000 /* Max. $rpipe() buffer. TODO: implement as gfd if you want more */
#define STRINGL 1024    /* String lengths.  Also, max script line length */
#define MAXPATH 1024    /* Max filename length (less or equal to STRINGL) */
#define MAXTOKEN 20     /* Maximum token or label length */

#define NVARS 286       /* a-z, a0-z9 == 26*11 */


extern char *optarg;
extern int optind, opterr;
FILE *filep;
char *script; // DO NOT MODIFY script!
int scriptspace;
bool ifres;
off_t lastpc,pc; /* program "counters" */
long resultcode=0; /* result code */
int waitfor_ignorecase=1;  /* no case sensitivity */
bool comecho=0; /* echo what's comin' in */  // TODO did I break this somewhen?
long senddelay=0; /* 0/100th second character delay for sending */
long number;  /* For getonearg() returning an long */
struct tagbstring *label; /* Names of labels for goto and gosub */
int labels;   /* Number of labels found in script */
long intvars[NVARS]; /* [a-z][0-9] integer variables */
bstring stringvars[NVARS]; /* $[a-z][0-9] string variables */
struct comgt_dev maindev;
struct tagbstring token;         /* For gettoken() returns */
char scriptfile[MAXPATH]; /* Script file name */
char scriptfilepath[MAXPATH]; /* temp storage for full path */
bool verbose=0; /* Log actions */
struct termios cons;  /* cons: stdout */
char msg[STRINGL]; /* Massage messages here */
int preturn,returns[MAXGOSUBS];
unsigned long hstart,hset;
bool lastcharnl=1; /* Indicate that last char printed from getonebyte
                               was a nl, so no new one is needed */



static void init_vars() {
    token = b_snew(.);
}


//"open com \"/dev/modem\"\nset com 38400n81\nset senddelay 0.05\nsend \"ATi^m\"\nget 2 \" ^m\" $s\nprint \"Response : \",$s,\"\\n\"\nget 2 \" ^m\" $s\nprint \"Response :\",$s,\"\\n\"\nget 2 \" ^m\" $s\nprint \"Response : \",$s,\"\\n\"\n\n";

/* Prototypes. */
void skipspaces(void);
#endif // #ifndef NOCODE
unsigned long htime(void);
void dormir(unsigned long microsecs);
void dotestkey(void);
void ext(long xtc);
void vmsg(const char *text);
void skipline(void);
void printwhere(void);
void writecom(struct comgt_dev* dev, bstring text);
int getonebyte(int fd);
void dodump(void);
void serror(char *text, int exitcode);
void getopen(void);
void getclose(void);
#define ALLCASE true
#define LOWERCASE false
int gettoken(bool);
void skiptoken(void);
long getvalue(void);
void getcomma(void);
bstring gethardstring(void);
bstring getstring(void);
unsigned long getdvalue(void);
void dolet(void);
int dowaitquiet(void);
int dowaitfor(void);
bool getonoroff(void);
int get_speed(struct comgt_dev* dev);
void setcom(struct comgt_dev* mdev);
void doset(void);
void dogoto(void);
void dogosub(void);
bool getonearg(void);
int getindex(void);
int getintindex(void);
int getstringindex(void);
void doget(void);
void doprint(int channel);
void doclose(void);
void doopen(void);
int doscript(void);
#ifndef NOCODE

#define NOCODE
#include "generic-fd.c"
#include "helper.c"
#include "xG.c"
#undef NOCODE



char *GTdevice[4] = {"/dev/noz2",
                     "/dev/ttyUSB2",
                     "/dev/modem",
                     (char*)0}; /* default device names to search for */



/* Returns hundreds of seconds */
unsigned long htime(void) {
  struct timeval timenow;
  gettimeofday(&timenow,NULL);
  return(100L*(timenow.tv_sec-hstart)+(timenow.tv_usec)/10000L-hset);
}

/* I use select() 'cause CX/UX 6.2 doesn't have usleep().
   On Linux, usleep() uses select() anyway.
*/
void dormir(unsigned long microsecs) {
  struct timeval timeout;
  timeout.tv_sec=microsecs/1000000L;
  timeout.tv_usec=microsecs-(timeout.tv_sec*1000000L);
  select(1,0,0,0,&timeout);
}

/* Tests for ENTER key */
void dotestkey(void) {
  fd_set fds;
  struct timeval timeout;
  timeout.tv_sec=0L;
  timeout.tv_usec=10000L;
  FD_ZERO(&fds);
  FD_SET(0,&fds);  /* Prepare to select() from stdin */
  resultcode=select(1,&fds,0,0,&timeout);
  if(resultcode) getchar();
}

/* Exit after resetting terminal settings */
void ext(long xtc) {
  tcsetattr(1, TCSADRAIN, &cons);
  destroy_gfd();
  exit(xtc);
}

/* Log message if verbose is on */
void vmsg(const char *text) {
  time_t t;
  char *ct;
  if(verbose) {
    if(lastcharnl==0) {
      fprintf(stderr,"\n");
      lastcharnl=1;
    }
    t=time(0);
    char* Test = malloc(3);
    ct=ctime(&t);
    fprintf(stderr,"comgt %c%c:%c%c:%c%c -> %s\n",
            ct[11],ct[12],ct[14],ct[15],ct[17],ct[18],
            text);
  }
}

/* Skip to next statement */
void skipline(void) {
  while(1) {
    if(pc>=scriptspace || script[pc]==0) return;
    if(script[pc]=='\n') {
      if(!pc || script[pc-1]!='\\') break;
    }
    pc++;
  }
  pc++; // Stand at \n, proceed to first char of next line
}

void printwhere(void) {
  int a,b,c;
  sprintf(msg,"@%04lld ",(long long)pc);
  a=pc;
  skipline();
  b=pc-1;
  pc=a;
  c=strlen(msg);
  for(;a<b;a++) msg[c++]=script[a];
  msg[c]=0;
  vmsg(msg);
}

static void printNext(void) {
  off_t i = pc, last_i = pc;
  while (i < pc+100) {
    if (script[i++] == '\n') {
      write(1, &(script[last_i]), i-last_i);
      last_i = i;
    }
  }
  if (last_i != i) {
    write(1, &(script[last_i]), i-last_i);
    puts(" (<- Begin of line)\n");
  }
}

/* Write a string to communication device */
void writecom(cdev dev, bstring text) {
  if(!senddelay) {
    if(!b_writeb(dev->fd, text)) {
      serror("Could not write to COM device",1);
    }
  }
  else {
    int a = b_blength(text);
    char* buf = text->data;
    for(;a;a--) {
      int res=b_writek(dev->fd,buf++,1);
      if(res!=1) {
        serror("Could not write to COM device",1);
      }
      dormir(senddelay);
    }
  }
}


// TODO: Performance for fds other than dev: don't read on single-byte-basis?

/* Gets a single byte from comm. device.
 *  Return -1 if none avail.
 *  Return -2 on error. */
int getonebyte(int fd) {
  fd_set rfds;
  int res;
  char ch;
  comecho = 1;
  struct timeval timeout;
  timeout.tv_sec=0L;
  timeout.tv_usec=10000;
  FD_ZERO(&rfds);
  FD_SET(fd, &rfds);
  res=select(fd+1,&rfds,NULL,NULL,&timeout);
//  printf("select(%d) -> %d\n", fd, res);
  if (res < 0) return -2;
  if (res == 0) return -1; /* Nada. */
  
  res=read(fd,&ch,1);
  if (res < 0) return 0 - (10 + errno);
  if (res == 0) return -2; // select told us there's something, but there's nothing -> looks like a closed pty
  if(res==1) {
    if(comecho) {
      if(ch=='\n') lastcharnl=1;
      else {
        if(ch!='\r') lastcharnl=0;
      }
      // TODO: Commented out?
      /*fputc(ch,stderr);*/
    }
    return(ch);
  }
  return(0);
}

void dodump(void) {
  char lmsg[STRINGL];
  int a,b,c;
  c=verbose;
  verbose=1;
  sprintf(lmsg,"-- Integer variables --"); vmsg(lmsg);
  for(a=0;a<26;a++) {
    for(b=0;b<11;b++) {
      if(intvars[b*26+a]) {
        sprintf(lmsg,"    = %ld",intvars[b*26+a]);
        lmsg[1]='a'+a;
        lmsg[2]=' ';
        if(b) lmsg[2]='0'+b-1;
        vmsg(lmsg);
      }
    }
  }
  sprintf(lmsg,"-- String variables --"); vmsg(lmsg);
  for(a=0;a<26;a++) {
    for(b=0;b<11;b++) {
      bstring str = stringvars[b*26+a];
      if(b_bsane(str) && b_blength(str)) {
        lmsg[0]='$';
        lmsg[1]='a'+a;
        lmsg[2]= b?'0'+b-1:' ';
        lmsg[3]=' ';
        lmsg[4]='=';
        lmsg[5]=' ';
        size_t remain = STRINGL - 6-2;
        
        int len = str->slen;
        if (len > STRINGL - 6-2-3) {
          len = STRINGL - 6-2-3;
          //TODO
        }
        memcpy(&(lmsg[6]), str->data, str->slen);
        lmsg[6+str->slen]=0;
        vmsg(lmsg);
      }
    }
  }
  verbose=c;
}

/* Report script errors and quit */
void serror(char *text, int exitcode) {
  verbose=1;
  
    char* Test = malloc(3);
  
  int a=0, line=1, lpc=0, finpc=0;
  while(script[a] != 0) {
    if(script[a++]=='\n') {
      if (a > pc) {
        finpc = a;
        break;
      }
      lpc = a;
      line++;
    }
    else if(a > pc+60) {
      finpc = a;
      break;
    }
  }
  //printf("lpc=%d, lastpc=%d, pc=%d, finpc=%d\n", lpc, lastpc, pc, finpc);
  char lmsg[STRINGL];
  sprintf(lmsg,"Error @%lld, line %d, %s. (%d)",(long long)pc,line,text,exitcode);
  vmsg(lmsg);
  
  strcpy(lmsg,"  ||  ");
  strNcpy(lmsg+6, &(script[lpc]), finpc-lpc);
  lmsg[6+finpc-lpc] = 0;
  vmsg(lmsg);
  
  a=pc;
  pc=lastpc;
  //printwhere();
  pc=a;
  while(pc!=0 && (script[pc]==' ' || script[pc]=='\t')) pc--; // Why?
  
  strcpy(lmsg,"  ||->");
  for(a=0;a<lastpc-lpc;a++) {
    if(script[lpc+a]=='\t') lmsg[a+6]='\t';
    else lmsg[a+6] = ' ';
  }
  for(;a<pc-lpc;a++) {
    if(script[lpc+a]=='\t') lmsg[a+6]='\t';
    else lmsg[a+6] = '=';
  }
  a+=6;
  lmsg[a++]='^';
  lmsg[a]=0;
  vmsg(lmsg);
  
  //dodump();
  
  //vmsg("-- Next lines would have been: --");
  //printNext();
  
  ext(exitcode);
}



#endif // #ifndef NOCODE
// always include
inline static void skipspaces_noglobal(char* script, size_t scriptspace, off_t* pc) {
  while(*pc < scriptspace && script[*pc]) {
    char ch=script[*pc];
    if(ch!=' '&&ch!='\t'&&ch!='\\') break;
    if(ch==' '||ch=='\t') (*pc)++;
    else {
      if(script[(*pc)+1]=='\n') (*pc)+=2;
	  else break;
    }
  }
}
#ifndef NOCODE

void skipspaces(void) {
  skipspaces_noglobal(script, scriptspace, &pc);
}


void getopen(void) {
  int a;
  a=pc;
  skipspaces();
  if(script[pc++]!='(') {
    pc=a;
    serror("Function requires open parenthesis",5);
  }
}

void getclose(void) {
  int a;
  a=pc;
  skipspaces();
  if(script[pc++]!=')') {
    pc=a;
    serror("Function requires close parenthesis",5);
  }
}

void skiptoken(void) {
  skipspaces();
  while(script[pc]!=' ' && script[pc]!='\t' && script[pc]!='\n' ) pc++;
  skipspaces();
}

/* Parse script[pc] to get next statement.  Resolve comments and labels... */
int gettoken(bool case_) {
  skipspaces();
  int startpc = pc;
  if(script[pc]=='"' || script[pc]=='\'')
    serror("Searching for cmd/func/number/comparator --  \"' not allowed here", 5);
  while (pc < scriptspace && (pc - startpc < MAXTOKEN)) {
    if(   script[pc]!='='
       && script[pc]!='<'
       && script[pc]!='>'
       && script[pc]!='!')
      break;
    pc++;
  }
  if (pc > startpc)  {
    token = b_sfromk_Ref(&script[startpc], pc-startpc);
    return(0);
  }
  if(script[pc]=='#') {
    token = b_sfromc_StaticRef("rem");
    pc++;
    return(0);
  }
  if(script[pc]=='/' && script[pc+1]=='/') {
    token = b_sfromc_StaticRef("rem");
    pc+=2;
    return(0);
  }
  if(script[pc]==':') {
    token = b_sfromc_StaticRef("label");
    pc++;
    return(0);
  }
  while(script[pc]!=' ' && script[pc]!='\n' &&
        script[pc]!='\t' && script[pc]!='(') {
    if(case_ == LOWERCASE && 'A'<=script[pc] && script[pc] <= 'Z')
      serror("In this version we only allow for lowercase tokens",5);
    pc++;
    if(pc-startpc==MAXTOKEN-1) serror("Token too long",5);
  }
  token = b_sfromk_Ref(&script[startpc], pc-startpc);
  if(b_seqc_Static(token,"then")) return(gettoken(case_)); /* Ignore then for if */
  return(0);
}

/* shitfaced recursive value parser. must write better one */
long getvalue(void) {
  unsigned long p=0;
  int goone=1;
  unsigned int a;
  int base,amode;
  bool found=false;
  //bool save_string_register = (string != NULL && string->slen >= 0);
  //if (string->mlen < 0)
  //  serror("Internal error - programmer should have used fully bstrlib managed string.", 9);
  
  skipspaces();
  if(script[pc]=='(') {
    pc++;
    p=getvalue();
    getclose();
    found=true;
  }
  else if(script[pc]==')') { // will remove unneccessary "else"s soon
    serror("???", 5);
    return(p);
  }
  else if(script[pc]=='"' || script[pc]=='\'' || script[pc]=='$' ) {
    serror("Did not expect a string",7);
  }
  else if( (script[pc]>='a' && script[pc]<='z' &&
            script[pc+1]>='a' && script[pc+1]<='z') ||
           (script[pc]>='A' && script[pc]<='Z' &&
            script[pc+1]>='A' && script[pc+1]<='Z') ) {
    gettoken(LOWERCASE);
    getopen();
    if(b_seqc_Static(token,"len")) {
      bstring string = getstring();
      p=b_blength(string);
      b_bdestroy(string);
    }
    else if(b_seqc_Static(token,"htime")) {
      p=htime();
    }
    else if(b_seqc_Static(token,"time")) {
      p=time(0);
    }
    else if(b_seqc_Static(token,"pid")) {
      p=getpid();
    }
    else if(b_seqc_Static(token,"ppid")) {
      p=getppid();
    }
    else if(b_seqc_Static(token,"verbose")) {
      p=verbose;
    }
    else if(b_seqc_Static(token,"isatty")) {
      p=isatty(getvalue());
    }
    else if(b_seqc_Static(token,"osnumber")) {
      int gfd=getfd(script, scriptspace, &pc);
      if (gfd < 0) serror("Wrong gfd index in osnumber() - expecting e.g. \"osnumber(!5)\"", 5);
      if (gfds[gfd].type == 0)
        p=-1;
      else
        p=gfds[gfd].fd;
    }
    else if(b_seqc_Static(token,"baud")) {
      int speed=get_speed(&maindev);
      if(maindev.high_speed && speed==115200)
        p=57600;
      else
        p=speed;
    }
    else if(b_seqc_Static(token,"access")) {
      bstring afile = getstring();
      getcomma();
      bstring mpars = getstring();
      if(mpars->slen<=0) serror("Missing access mode[s]",5);
      amode=0;
      for(a=0;a<b_blength(mpars);a++) {
        switch(b_ratb(mpars,a)) {
          case 'R' :
          case 'r' : amode|=R_OK; break;
          case 'W' :
          case 'w' : amode|=W_OK; break;
          case 'X' :
          case 'x' : amode|=X_OK; break;
          case 'F' :
          case 'f' : amode|=F_OK; break;
          default : serror("Access modes are r,w,x, and f",5);
        }
      }
      b_bdestroy(mpars);
      p=access(afile->data,amode);
      b_bdestroy(afile);
    }
    else if(b_seqc_Static(token,"val") || b_seqc_Static(token,"atol") || b_seqc_Static(token,"int")) {
      p=atol(b_cfromb_CopyRepl_(getstring()));
    }
    else serror("Unknown Integer function",5);
    getclose();
    found=true;
  }
  else if(script[pc]=='v' && script[pc+1] == '3' && script[pc+2] == 'r') {
    pc+=3;
    p=intvars['r'-'a']+1; // Ensure result is (v3)r != r
    found=true;
  }
  if(script[pc]=='%') { // else if would also work // TODO: better error handling?
    pc++;
    p=resultcode;
    found=true;
  }
  while(goone) {
    // TODO: either this gets a big loop over everything or it gets a 
    // strictly recursive approach. In both cases it makes sense to 
    // distinguish between operator and operand (mainly because of "%"
    // resultcode operand) btw. modulo operator is missing.
    bool notfound=false;
    if(script[pc]=='+') {
      pc++;
      p+=getvalue();
    }
    else if(script[pc]=='-') {
      pc++;
      p-=getvalue();
    }
    else if(script[pc]=='^') {
      pc++;
      p^=getvalue();
    }
    else if(script[pc]=='&') {
      pc++;
      p&=getvalue();
    }
    else if(script[pc]=='|') {
      pc++;
      p|=getvalue();
    }
    else if(script[pc]=='*') {
      pc++;
      p*=getvalue();
    }
    else if(script[pc]=='/') {
      pc++;
      a=getvalue();
      if(a==0) serror("Division by zero",6);
      p/=a;
    }
    else if((script[pc]>='a' && script[pc]<='z') ||
            (script[pc]>='A' && script[pc]<='Z') ) {
      p=intvars[getintindex()];
      found=true;
    }
    else if(script[pc]>='0' && script[pc]<='9') {
      base=10;
      if(script[pc]=='0') {
        base=8;
        pc++;
        if(script[pc]=='x' || script[pc]=='X') {
          base=16;
          pc++;
        }
        if(script[pc]=='%') {
          base=2;
          pc++;
        }
      }
      while((script[pc]>='0' && script[pc]<='9') ||
            (script[pc]>='a' && script[pc]<='f') ||
            (script[pc]>='A' && script[pc]<='F')) {
        if(script[pc]>='a' && script[pc]<='f') {
          p=p*base+script[pc++]-'a'+10;
        }
        else if(script[pc]>='A' && script[pc]<='F') {
          p=p*base+script[pc++]-'A'+10;
        }
        else {
          p=p*base+script[pc++]-'0';
        }
      }
    }
    else {
      if(!found)
        // empty line
        serror("Searched for number, but didn't find any",5);
      notfound=true;
      goone=0;
    }
    if (!notfound) found=true;
  }
  return(p);
}

void getcomma(void) {
  skipspaces();
  if(script[pc++]!=',') serror("Comma expected",5);
}

bstring gethardstring(void) {
  int a=0;
  skipspaces();
  while(script[pc+a]!=0 && script[pc+a]!=' ' && script[pc+a]!='\t' &&
        script[pc+a]!='\n') {
    a++;
  }
  bstring retv = b_bfromk(&script[pc], a);
  pc += a;
  return retv;
}

/* Parse a string from script[pc] */
bstring getstring(void) {
  
  FILE *fp;
  time_t t;
  //unsigned int a,b;
  int c,p=0;
  unsigned char ch,match;
  bstring string=b_bnew(.);
  skipspaces();
  if(script[pc]!='"' && script[pc]!='\'' && script[pc]!='$' ) {
    serror("Expected a string",7);
  }
  while(script[pc]!=' ' && script[pc]!='\t' && script[pc]!='\n' && script[pc]!=','  && script[pc]!=')' && script[pc]!='=' && script[pc]!='<' && script[pc]!='>' && script[pc]!='!' ) {
    if(script[pc]=='+') pc++;
    skipspaces();
    if( (script[pc]=='$' && script[pc+1]>='a' && script[pc+1]<='z' &&
         script[pc+2]>='a' && script[pc+2]<='z') ||
        (script[pc]=='$' && script[pc+1]>='A' && script[pc+1]<='Z' &&
         script[pc+2]>='A' && script[pc+2]<='Z') ) {
      pc++;
      gettoken(LOWERCASE);
      getopen();
      if(b_seqc_Static(token,"version")) {
        b_bcatc_Static(string,"comgt version ");
        b_bcatc       (string,comgt_version);
        b_bcatc_Static(string," Copyright ");
        b_bcatc_Static(string,AUTHORS);
        b_bcatc_Static(string," (c) ");
        b_bcatc_Static(string,YEARS);
      }
      else if(b_seqc_Static(token,"versionnumber")) {
        b_bcatc(string,comgt_version);
      }
      else if(b_seqc_Static(token,"time")) {
        t=time(0);
        b_bcatc(string,ctime(&t));
        string->slen--;
      }
      else if(b_seqc_Static(token,"rpipe")) {
        bstring cmd = getstring();
        fflush(stdout);
        // Dangerous - using string->data, but bstrlib normally add 0 to end.
        if((fp=popen(cmd->data,"r"))==NULL) serror("Could not popen!",4);
        b_bdestroy(cmd);
        b_bmemgrow(string, b_blength(string) + RPIPE_OUTPUT_MAXL + 1);
        if (!b_bsane(string))
          serror("[getstring] Internal string problem", 9);
        size_t res = fread(&string->data[string->slen],1,RPIPE_OUTPUT_MAXL,fp);
        pclose(fp);
        // Don't set error code (probably best to be reserved to commands).
        if (res < 0) {
          b_bcatc_Static(string, "COMGT $rpipe() read error!!");
        }
        else {
          string->slen += res;
          if (string->data[string->slen-1] == 0x0a)
            string->slen--; // Remove trailing newline
          string->data[string->slen] = 0; // Finalize in bstrlib manner
        }
      }
      else if(b_seqc_Static(token,"env")) {
        bstring varname = getstring();
        // Dangerous - using string->data, but bstrlib normally add 0 to end.
        char* envvar = getenv(varname->data);
        b_bdestroy(varname);
        if(envvar)
          b_bcatc(string, envvar);
      }
      else if(b_seqc_Static(token,"hms")) {
        long sec,min,hour;
        sec=getvalue();
        min=sec/60L;
        sec-=min*60L;
        hour=min/60L;
        min-=hour*60L;
        b_bcatc_Format(string,"%02ld:%02ld:%02ld",hour,min,sec);
      }
      else if(b_seqc_Static(token,"name")) {
        int gfd=getfd(script, scriptspace, &pc);
        if (gfd < 0) serror("Wrong gfd index in $name() - expecting e.g. \"$name(!5)\"", 5);
        b_bcatc(string,gfd_getname(gfd));
      }
      else if(b_seqc_Static(token,"type")) {
        int gfd=getfd(script, scriptspace, &pc);
        if (gfd < 0) serror("Wrong gfd index in $type() - expecting e.g. \"$type(!5)\"", 5);
        b_bcatr(string,gfds[gfd].type);
      }
      else if(b_seqc_Static(token,"ptsname")) {
        int fdrequest=getfd(script, scriptspace, &pc);
        if (fdrequest<0) serror("$ptsname() requires a gfd number.", 7);
        if (gfds[fdrequest].type == 0)   serror("$ptsname(): this fd is closed.", 7);
        if (gfds[fdrequest].type != 'p') serror("$ptsname(): this is no pts.", 7);
        char* name=ptsname(gfds[fdrequest].fd);
        if (name==NULL) serror("$ptsname(): syscall returned NULL.", 8);
        b_bcatc(string,name);
      }
      else if(b_seqc_Static(token,"dev")) {
        b_bcatb(string,maindev.filename);
      }
      else if(b_seqc_Static(token,"cwd")) {
        char buf[STRINGL];
        getcwd(buf,STRINGL);
        b_bcatc(string, buf);
      }
      else if(b_seqc_Static(token,"baud")) { // ToDo: remove this in favour of $str(baud())
        int speed=get_speed(&maindev);
        if(speed==57600 && maindev.high_speed)
          speed=115200;
        char buf[10];
        snprintf(buf, sizeof(buf), "%d", speed);
        b_bcatc(string,buf);
      }
      else if(b_seqc_Static(token,"txt2gsm7bit")) {
        int lbefore = b_blength(string);
        char* newstring = &(string->data[lbefore]); // zero-cost expression
        
        // It would be very memory-efficient if getstring would operate
        // on newstring (and some conditions are met, e. g. strings are short
        // and/or expanding large strings is cheap
        bstring in = getstring();
        if (!b_bsane(in)) serror("[getstring] Internal string problem", 9);
        int txtlen = b_blength(in);
        
        // Data will be "shrunk".
        b_bmemgrow(string, lbefore + (txtlen*7+7)/8 + 1);
        if (!b_bsane(string)) serror("[getstring] Internal string problem", 9);
        
        txtlen = txt2gsm7bit(newstring, in->data, txtlen);
        newstring[txtlen] = 0;
        string->slen += txtlen;
      }
      else if(b_seqc_Static(token,"gsm7bit2txt")) {
        int lbefore = b_blength(string);
        char* newstring = &(string->data[lbefore]); // zero-cost expression
        
        // It would be very memory-efficient if getstring would operate
        // on newstring (and some conditions are met, e. g. strings are short
        // and/or expanding large strings is cheap
        bstring in = getstring();
        if (!b_bsane(in)) serror("[getstring] Internal string problem", 9);
        int txtlen = b_blength(in);
        
        // Data will be "expanded".
        b_bmemgrow(string, lbefore + (txtlen*8)/7 + 1);
        if (!b_bsane(string)) serror("[getstring] Internal string problem", 9);
        
        txtlen = gsm7bit2txt(newstring, in->data, txtlen);
        newstring[txtlen] = 0;
        string->slen += txtlen;
      }
      else if(b_seqc_Static(token,"txt2pdu7bit")) {
        // Data will be "expanded".
        // [pdu7bit <-> txt engine was carefully written to operate in-place, and now we just don't use it...]
        int lbefore = b_blength(string);
        char* newstring = &(string->data[lbefore]); // zero-cost expression
        
        bstring in = getstring();
        if (!b_bsane(in)) serror("[getstring] Internal string problem", 9);
        int txtlen = b_blength(in);
        int tgtlen = (txtlen*2*7+7)/8;
        
        b_bmemgrow(string, string->slen + tgtlen + 1);
        if (!b_bsane(string)) serror("[getstring] Internal string problem", 9);
        
        txtlen = txt2pdu7bit(newstring, in->data, txtlen);
        b_bdestroy(in);
        if (txtlen < 0) serror("[7bit] Internal string problem", 9);
        if (txtlen != tgtlen) serror("[7bit] Output length mismatch", 9);
        newstring[txtlen] = 0;
        string->slen += txtlen;
        
        // TODO: Rename functions to txt2gsm7bithex / gsm7bithex2txt
      }
      else if(b_seqc_Static(token,"pdu7bit2txt")) {
        // Data will be "shrunk".
        int lbefore = b_blength(string);
        char* newstring = &(string->data[lbefore]); // zero-cost expression
        
        bstring in = getstring();
        int txtlen = b_blength(in);
        
        // Too lazy to calculate - output will be at most input byte size
        b_bmemgrow(string, string->slen + txtlen + 1);
        if (!b_bsane(string)) serror("[getstring] Internal string problem", 9);
        
        txtlen = pdu7bit2txt(newstring, in->data, txtlen);
        b_bdestroy(in);
        if (txtlen < 0) serror("[7bit] Internal string problem", 9);
        
        newstring[txtlen] = 0;
        string->slen += txtlen;
      }
      else if(b_seqc_Static(token,"str") || b_seqc_Static(token,"ltoa") || b_seqc_Static(token,"val")) {
        b_bcatc_Format(string,"%ld",getvalue());
      }
      else if(b_seqc_Static(token,"hexu")) {
        b_bcatc_Format(string,"%lX",getvalue());
      }
      else if(b_seqc_Static(token,"hex")) {
        b_bcatc_Format(string,"%lx",getvalue());
      }
      else if(b_seqc_Static(token,"oct")) {
        b_bcatc_Format(string,"%lo",getvalue());
      }
      else if(b_seqc_Static(token,"tolower")) {
        bstring in = getstring();
        for(off_t a=0;a<b_blength(in);a++) {
          char ch=b_ratb(in,a);
          // TODO: Let libc function do this with locale
          if(ch>='A' && ch<='Z' ) in->data[a]=ch-'A'+'a';
        }
        b_bcatb(string, in);
        b_bdestroy(in);
      }
      else if(b_seqc_Static(token,"toupper")) {
        bstring in = getstring();
        for(off_t a=0;a<b_blength(in);a++) {
          char ch=b_ratb(in,a);
          if(ch>='a' && ch<='z' ) in->data[a]=ch-'a'+'A';
        }
        b_bcatb(string, in);
        b_bdestroy(in);
      }
      else if(b_seqc_Static(token,"basename")) {
        bstring full = getstring();
        off_t l = b_blength(full);
        if(l>1 && full->data[l-1]=='/') l--;
        if(l) {
          off_t a=l;
          for(;a;a--) {
            if(full->data[a-1]=='/' || full->data[a-1]=='\\') break;
          }
          b_bcatk(string, &full->data[a], l-a);
        }
        b_bdestroy(full);
      }
      else if(b_seqc_Static(token,"dirname")) {
        bstring full = getstring();
        off_t l = b_blength(full);
        if(l>1 && full->data[l-1]=='/') l--;
        if(l) {
          for(;l;l--) {
            if(full->data[l-1]=='/' || full->data[l-1]=='\\') break;
          }
          if(l)
            b_bcatk(string, full->data, l-1);
        }
        b_bdestroy(full);
      }
      else if(b_seqc_Static(token,"script")) {
        b_bcatc(string, scriptfile);
      }
      else if(b_seqc_Static(token,"right")) {
        bstring in = getstring();
        getcomma();
        int nright=getvalue();
        int originallen = b_blength(in);
        int start = 0, stop = originallen;
        if(nright>0 &&  nright>originallen) start = -1;
        if(nright<0 && -nright>originallen) start = -1;
        if(start == -1) serror("String is shorter than second argument",7);
        if(nright<0) start =           0 - nright;
        else         start = originallen - nright;
        b_bcatk(string, &in->data[start], stop-start);
        b_bdestroy(in);
      }
      else if(b_seqc_Static(token,"left")) {
        bstring in = getstring();
        int originallen = b_blength(in);
        getcomma();
        int nleft=getvalue();
        int start = 0, stop = originallen;
        if(nleft>0 &&  nleft>originallen) start = -1;
        if(nleft<0 && -nleft>originallen) start = -1;
        if(start == -1) serror("String is shorter than second argument",7);
        if(nleft<0) stop =  nleft+originallen;
        else        stop =  nleft;
        b_bcatk(string, &in->data[start], stop-start);
        b_bdestroy(in);
      }
      else if(b_seqc_Static(token,"mid")) {
        bstring in = getstring();
        int originallen = b_blength(in);
        getcomma();
        int start=getvalue();
        getcomma();
        int len=getvalue();
        if(start < 0 || len < 0) 
          serror("Negative input not possible here",7);
        if(start > originallen || start + len > originallen)
          serror("String index out of bounds",7);
        b_bcatk(string, &in->data[start], len);
        b_bdestroy(in);
      }
      else if(b_seqc_Static(token,"sex")) {
        b_bcatc_Static(string,"You're a naughty guy, you!");
      }
      else serror("Invalid string funtion",5);
      getclose();
    }
    else if(script[pc]=='$') {
      b_bcatb(string,stringvars[getstringindex()]);
    }
    else if(script[pc]=='"' || script[pc]=='\'') {
      int originalpos=pc; // For error reporting.
      match=script[pc++];
      while(script[pc]!=match) {
        ch=script[pc++];
        if(ch=='\n') {
          pc--; lastpc=originalpos;
          serror("Quoted string floats over end-of-line - please fix with \" or make multiline explicit with \\<newline>!", 5);
        }
        if(ch==0 || pc>=scriptspace) {
          pc--; lastpc=originalpos;
          serror("Quoted string must be closed till end of file!", 5);
        }
        if(ch=='\\') {
          if(script[pc]<='7' && script[pc]>='0' &&
            script[pc+1]<='7' && script[pc+1]>='0' ) {
            ch=0;
            while(script[pc]>='0' && script[pc]<='7') {
              ch=8*ch+script[pc++]-'0';
            }
          }
          else {
            switch(script[pc]) {
              case 'T' :
              case 't' : ch=9;  break;
              case 'R' :
              case 'r' : ch=13; break;
              case 'N' :
              case 'n' : ch=10; break;
              case 'B' :
              case 'b' : ch=8;  break;
              case 'F' :
              case 'f' : ch=12; break;
              case 'e' : ch=0x1b; break; // Escape, \033, ^[
              case '"' :
              case '^' :
              case '\'' :
              case '\n' :
              case '\\' : ch=script[pc]; break;
              default :  serror("Malformed escaped character",5);
            }
            pc++;
          }
        }
        else if(ch=='^') {
          ch=script[pc];
          if(ch!='^' && ch!='"' && ch!='\'' && ch!='\\' ) {
            ch=ch&31; /* Control char */
          }
          pc++;
        }
        b_bcatr(string, ch);
      }
      pc++; /* " quote */
    }
    else if ('a' <= script[pc] && script[pc] <= 'z') {
      int idx = getintindex();
      char buf[10];
      snprintf(buf, 10, "%ld", intvars[idx]);
      b_bcatc(string, buf);
    }
    else {
      // TODO: This is not good. Taking any char, putting it onto the string??
      b_bcatr(string, script[pc++]);
    }
  }
  return string;
}

/* Get a value, multiply by a hundred (for time values) */
unsigned long getdvalue(void) {
  float f;
  gettoken(LOWERCASE);
  skipspaces();
  sscanf(token.data,"%f",&f);
  f+=0.00001; /* Rounding errors */
  return(100.0*f);
}

void dolet(void) {
  int index;
  bool svar=0;
  skipspaces();
  if(script[pc]=='$') {
    svar=1;
    index=getstringindex();
  }
  else index=getintindex();
  skipspaces();
  gettoken(ALLCASE);
  if(b_seqc_Static(token,"=")<=0) serror("Bad LET assignment, '=' missing",5);
  skipspaces();
  if(svar) {
    bstring newvalue=getstring();
    if (stringvars[index]) b_bdestroy(stringvars[index]);
    stringvars[index] = newvalue;
  }
  else {
    intvars[index]=getvalue();
  }
}

/* See documentation for doXXX() functions */
int dowaitquiet(void) {
  unsigned long timeout,timequiet,quiet,now;
  int c,quit;
  timeout=htime()+getdvalue();
  quiet=getdvalue();
  timequiet=htime()+quiet;
  int ptyfd=getfd(script, scriptspace, &pc);
  quit=1;
  while(quit==1) {
    now=htime();
    if (ptyfd >= 0)
      c=getonebyte(gfds[ptyfd].fd);
    else
      c=getonebyte(maindev.fd);
    if(c>=0) timequiet=now+quiet;
    if(now>=timequiet) quit=0;
    if(now>=timeout) quit=255;
  }
  return(quit);
}

#define WAITFOR_MAX_EXPRESSIONS 20
#define xstr(s) str(s)
#define str(s) #s
int dowaitfor(void) {
  bstring expressions[WAITFOR_MAX_EXPRESSIONS];
  char strings[20][80];
  char buffer[128];
  unsigned long timeout;
  unsigned int a, b=0;
  int c;
  buffer[127]=0;
  skipspaces();
  int funcpc=lastpc;
  lastpc=pc-1;
  timeout=htime()+getdvalue();
  lastpc=pc+1;
  int ptyfd=getfd(script, scriptspace, &pc);
  lastpc=pc+1;
  while(script[pc]==',' || script[pc]=='$' || script[pc]=='"' ||
        script[pc]=='\'' ) {
    if(script[pc]==',') pc++;
    skipspaces();
    expressions[b] = getstring();
    if(waitfor_ignorecase) b_btolower(expressions[b]);
    skipspaces();
    b++;
    if(b>=WAITFOR_MAX_EXPRESSIONS)
      serror("Can only wait for up to " xstr(WAITFOR_MAX_EXPRESSIONS) " expressions", 5);
  }
  lastpc=funcpc;
  while(htime()<timeout) {
    // TODO: performance (not move 126 bytes for every byte ...)
    if (ptyfd >= 0)
      c=getonebyte(gfds[ptyfd].fd);
    else
      c=getonebyte(maindev.fd);
    //printf("Byte \"%c\" ",c);
    if(c>=0) {
      if(waitfor_ignorecase) {
        if(c>='A' && c<='Z') c=c-'A'+'a';
      }
      for(a=0;a<127;a++) buffer[a]=buffer[a+1]; //shuffle down
      buffer[126]=c;
      for(int i=0; i<b; i++) {
        c=b_blength(expressions[i]);
        if (b_beqc(expressions[i],&buffer[127-c])){
          return(i);
        }
      }
    }
  }
  return(-1);
}

/* Parse script for "on" or "off" which are tokens, not strings */
bool getonoroff(void) {
  int a,b;
  b=pc;
  gettoken(LOWERCASE);
  if(b_seqc_Static(token,"on")) return(1);
  if(b_seqc_Static(token,"off")) return(0);
  pc=b;
  a=getvalue();
  if(a!=0 && a!=1) serror("Bad value (should be on or off, 1 or 0.)",5);
  return(a);
}

int int2baud(int i) {
  switch(i) {
    case 0:      return B0;
    case 50:     return B50;
    case 75:     return B75;
    case 110:    return B110;
    case 150:    return B150;
    case 300:    return B300;
    case 600:    return B600;
    case 1200:   return B1200;
    case 2400:   return B2400;
    case 4800:   return B4800;
    case 9600:   return B9600;
    case 19200:  return B19200;
    case 38400:  return B38400;
    case 57600:  return B57600;
    case 115200: return B115200;
	case 230400: return B230400;
#ifdef B460800
    case 460800: return B460800;
#endif
    default: serror("Invalid baudrate",1);
  }
  return -555; // Never reaching this
}

int baud2int(int b) {
  switch(b) {
    case B0:      return 0;
    case B50:     return 50;
    case B75:     return 75;
    case B110:    return 110;
    case B150:    return 150;
    case B300:    return 300;
    case B600:    return 600;
    case B1200:   return 1200;
    case B2400:   return 2400;
    case B4800:   return 4800;
    case B9600:   return 9600;
    case B19200:  return 19200;
    case B38400:  return 38400;
    case B57600:  return 57600;
    case B115200: return 115200;
	case B230400: return 230400;
#ifdef B460800
    case B460800: return 460800;
#endif
    default: serror("Invalid baudrate",1);
  }
  return -555; // Never reaching this
}

int get_speed(cdev dev) { // returns plain integer, eg. 57600
  int ri = cfgetispeed(&dev->work_termio);
  int ro = cfgetospeed(&dev->work_termio);
  if(ri!=ro) {
    bstring msg = b_bfromc_Static("You asked for speed, but ispeed and ospeed differ ");
	b_bcatc_Format(msg, "(%d-%d), so giving you ispeed.", ri, ro);
    vmsg(msg->data);
  }
  return baud2int(ri);
}
void set_speed(cdev dev, int speed) { // speed = plain integer, eg. 57600
  cfsetspeed(&dev->work_termio, int2baud(speed));
}
bool get_clocal(cdev dev) {
  return dev->work_termio.c_cflag &= CLOCAL ? true : false;
}
void set_clocal(cdev dev, bool clocal_on) {
  if (clocal_on) dev->work_termio.c_cflag &= ~CLOCAL;
  else           dev->work_termio.c_cflag |=  CLOCAL;
}
void set_size(cdev dev, char sz) {
  int bits=0;
  switch(sz) {
    case '5' : bits=CS5; break;
    case '6' : bits=CS6; break;
    case '7' : bits=CS7; break;
    case '8' : bits=CS8; break;
    default : serror("Bits can only be 5, 6, 7, or 8",5);
  }
  dev->work_termio.c_cflag &= ~CSIZE;
  dev->work_termio.c_cflag |= bits;
}
void set_parity(cdev dev, char parity) { // Only set needed, no get
  int p=0;
  switch(parity) {
    case 'n': break;
    case 'e': p=PARENB; break;
    case 'o': p=PARENB|PARODD; break;
    default : serror("Parity can only ben E, N, or O",5);
  }
  dev->work_termio.c_cflag &= ~(PARENB|PARODD);
  dev->work_termio.c_cflag |= p;
}
void set_stopbits(cdev dev, char s) { // Only set needed, no get
  int stopbits=0;
  switch(s) {
    case '1': break;
    case '2': stopbits=CSTOPB; break;
    default : serror("Stop bits can only be 1 or 2",5);
  }
  dev->work_termio.c_cflag &= ~CSTOPB;
  dev->work_termio.c_cflag |= stopbits;
}

#endif // #ifndef NOCODE

void init_comgt_dev_struct(struct comgt_dev* cdev)
#ifndef NOCODE
{
  if (!cdev) return;
  cdev->filename=NULL;
  cdev->fd=-1;
  cdev->high_speed=false;
  set_speed   (cdev, 0);
  set_size    (cdev, '8');
  set_parity  (cdev, 'n');
  set_stopbits(cdev, '1');
  set_clocal  (cdev, 0);
}
#endif // #ifndef NOCODE
;


#ifndef NOCODE

void setcom(cdev dev) {
  if(dev==NULL) serror("Can't set com property for NULL dev!", 9);
  if (tcsetattr(dev->fd, TCSADRAIN, &dev->work_termio) < 0) {
    serror("Can't ioctl set device",1);
  }
}

void doset(void) {
  struct termios console;
  int a,b;
  gettoken(LOWERCASE);
  if(b_seqc_Static(token,"echo")) {
    a=0;
    if(getonoroff()) a=ECHO|ECHOE;
    if(tcgetattr(0, &console)<0) {
      serror("Can't ioctl FD zero!\n",2);
    }
    console.c_lflag &= ~(ECHO | ECHOE);
    console.c_lflag |= a;
    tcsetattr(0, TCSADRAIN, &console);
  }
  else if(b_seqc_Static(token,"senddelay")) {
    senddelay=10000L*getdvalue();
  }
  else if(b_seqc_Static(token,"clocal")) {
    set_clocal(&maindev, getonoroff());
    setcom(&maindev);
  }
  else if(b_seqc_Static(token,"umask")) {
    umask(getvalue()&0777);
  }
  else if(b_seqc_Static(token,"verbose")) {
    verbose=getonoroff();
  }
  else if(b_seqc_Static(token,"comecho")) {
    comecho=getonoroff();
  }
  else if(b_seqc_Static(token,"ignorecase")) {
    waitfor_ignorecase=getonoroff();
  }
  else if(b_seqc_Static(token,"com")) {
    cdev cdev=&maindev;
    skipspaces();
    if(script[pc]=='$' || script[pc]=='\'' || script[pc]=='"') {
      bstring arg = getstring();
      token = b_bdemarshal(arg);
    }
    else gettoken(LOWERCASE);
    a=0;
    b=0;
    while(b_rats(token, b)>='0' && b_rats(token, b)<='9') {
      a=10*a+b_rats(token, b)-'0';
      b++;
    }
    if (a==115200 && cdev->high_speed)
      set_speed(cdev, 57600); // Make speed lower when high_speed is on
    else
      set_speed(cdev, a);
    if(b_rats(token, b)) {
      set_parity(cdev, b_rats(token, b));
      b++;
      if(b_rats(token, b)) {
        set_size(cdev, b_rats(token, b));
        b++;
        if(b_rats(token, b)) {
          set_stopbits(cdev, b_rats(token, b));
        }
      }
    }
    setcom(cdev); // commit all changes
  }
  else serror("'set': don't know given configuration variable. Variable assignment works like \"let a=3\"",7);
}

void dogoto(void) {
  int a,originalpos;
  originalpos=pc;
  gettoken(ALLCASE);
  for(a=0;a<labels;a++) {
    if(b_seqs(token,label[a])) break;
  }
  if(a>=labels) {
    pc=originalpos;
    bstring msg = b_bnew(.);
    b_bcatc_Static(msg, "Label \"");
    b_bcats       (msg, token);
    b_bcatc_Static(msg, "\" not found");
    char* cstr = b_cfromb_CopyRepl_(msg);
    b_bdestroy(msg);
    serror(cstr,5);
  }
  
  pc=label[a].data - (unsigned char*)script;
  skipline();
}

void dogosub(void) {
  int a;
  if(preturn==MAXGOSUBS) serror("Reached maximum GOSUB depth",3);
  a=pc;
  gettoken(ALLCASE);
  returns[preturn++]=pc;
  pc=a;
  dogoto();
}

/* Gets arguments and returns 0 for a string, 1 for an int. Used with if */
bool argisnumber(void) {
  if(script[pc]=='"' || script[pc]=='\'' || script[pc]=='$' )
    return 0;
  return 1;
}

static void comparenumbers() {
    int leftint=getvalue();
    gettoken(ALLCASE);
    skipspaces();
    if(!argisnumber())
      serror("Cannot compare int (left) to string (right)",7);
    int rightint=getvalue();
    if(b_seqc_Static(token,"<")) {
      if(leftint<rightint) ifres=1;
    }
    if(b_seqc_Static(token,"<=")) {
      if(leftint<=rightint) ifres=1;
    }
    else if(b_seqc_Static(token,"=")) {
      if(leftint==rightint) ifres=1;
    }
    else if(b_seqc_Static(token,">")) {
      if(leftint>rightint) ifres=1;
    }
    else if(b_seqc_Static(token,">=")) {
      if(leftint>=rightint) ifres=1;
    }
    else if(b_seqc_Static(token,"!=") || b_seqc_Static(token,"<>")) {
      if(leftint!=rightint) ifres=1;
    }
}

static void comparestrings() {
    bstring left = getstring();
    gettoken(ALLCASE);
    bstring tokencopy = b_bfroms(token);
    skipspaces();
    if(argisnumber())
      serror("Cannot compare string (left) to int (right)",7);
    bstring right = getstring();
    
    if(b_beqc_Static(tokencopy,"<")) {
      if(b_bcmpb(left,right)<0) ifres=1;
    }
    if(b_beqc_Static(tokencopy,"<=")) {
      if(b_bcmpb(left,right)<=0) ifres=1;
    }
    if(b_beqc_Static(tokencopy,"=")) {
      if(b_bcmpb(left,right)==0) ifres=1;
    }
    if(b_beqc_Static(tokencopy,">")) {
      if(b_bcmpb(left,right)>0) ifres=1;
    }
    if(b_beqc_Static(tokencopy,">=")) {
      if(b_bcmpb(left,right)>=0) ifres=1;
    }
    else if(b_beqc_Static(tokencopy,"!=") || b_beqc_Static(tokencopy,"<>")) {
      if(b_bcmpb(left,right)!=0) ifres=1;
    }
    b_bdestroy(left);
    b_bdestroy(tokencopy);
    b_bdestroy(right);
}



int getindex(void) {
  int index;
  index=script[pc++];
  if(index>='A' && index<='Z') index=index-'A'+'a';
  if(index>'z' || index<'a') serror("Malformed variable name",7);
  index=index-'a';
  if(script[pc]>='0' && script[pc]<='9') {
    index=index+(1+script[pc++]-'0')*26;
  }
  return(index);
}

/* Parse script to find integer variable index */
int getintindex(void) {
  skipspaces();
  if(script[pc]=='$') serror("Integer variable expected",7);
  return(getindex());
}

/* Parse script to find string variable index and allocate memory for storage
   as needed. */
int getstringindex(void) {
  int index;
  skipspaces();
  //printf("script[%d]='%c' (getstringindex)\n", pc, script[pc]);
  if(script[pc++]!='$') serror("String variable expected",7);
  return getindex();
}




// Result:
// 2 = Very long line, didn't read end.
// 1 = Timeout, maybe some bytes read.
// 0 = OK or nothing read
// -2 and lower = Error
int doget_loop(bstring * newdata, int fd, const char* terminators, unsigned long timeout, bool allowempty) {
  unsigned int a;
  if (*newdata != NULL)
    serror("[gethardstring] Cowardly refusing to overwrite existing string", 9);
  *newdata = b_bnew(.);
  int b=0,c;
  int goahead=1;
  
  while(goahead) {
    if (timeout && htime()>timeout) return 1; // Timeout
    c=getonebyte(fd);
//		printf("GOT: %c (0x%02x) from fd %d (gfd ?)\n", c, ((uint8_t)c), fd);
    if (c <= -2) return c; // Error
    if (c == -1) continue;
    
    for(a=0;a<strlen(terminators);a++) {
      if(c==terminators[a]) goahead=0;
    }
    if(goahead==0 && b==0 && !allowempty)
      goahead=1; /* Ignore terminators if nothing yet */
    else if(goahead) {
      b++; // Indicator that something was received. Could also be boolean.
      if (b_bcatr(*newdata, c) != BSTR_OK)
        return -2;
    }
  }
  
  return 0; // OK
}


void doget(void) {
  
  unsigned long timeout;
  timeout=htime()+getdvalue();
  
  int fd = maindev.fd;
  int ptyfd=getfd(script, scriptspace, &pc);
  if (ptyfd >= 0) fd = gfds[ptyfd].fd;
  
  
  bstring terminators = getstring();
  
  int index=getstringindex();
  
  bstring in = NULL;
  resultcode = doget_loop(&in, fd, terminators->data, timeout, false);
  b_bdestroy(terminators);
  if (stringvars[index]) b_bdestroy(stringvars[index]);
  stringvars[index] = in;
}

void doprint(int channel) {
  int16_t ptyfd=getfd(script, scriptspace, &pc);
  skipspaces();
  if (ptyfd >= 0) channel=5;
  char* tmp=malloc(8);
  struct tagbstring bmsg = (struct tagbstring)
      {.data=tmp, .slen=0, .mlen=8}; // TODO: make this a bstringlib function
  while(script[pc]!=' ' && script[pc]!='\t' && script[pc]!='\n') {
    if(pc>=scriptspace)
      serror("print command: somehow I'm reading over code boundary!", 5);
    if(script[pc]==',') pc++;
    else {
      if(script[pc]=='"' || script[pc]=='\'' || script[pc]=='$' ) {
        bstring expr=getstring();
        b_bcatb(&bmsg, expr);
        b_bdestroy(expr);
      }
      else {
        b_bcatc_Format(&bmsg, "%ld",getvalue());
      }
    }
  }
  //printf("Shall print '%s' @%d(!%d).\n", msg, channel, ptyfd); fflush(stdout);
  switch(channel) {
    case 1: write(1, bmsg.data, bmsg.slen); fflush(stdout); break;
    case 2: write(2, bmsg.data, bmsg.slen); break;
    case 3:
      if(bmsg.data[bmsg.slen-1]=='\n') bmsg.data[bmsg.slen---1]=0;
      vmsg(bmsg.data);
      break;
    case 4:
      if(filep==NULL) serror("File not opened",4);
      fwrite(bmsg.data, bmsg.slen, 1, filep);
      break;
    case 5:
      resultcode=gfd_write(ptyfd, &bmsg);
      break;
  }
  b_sdestroy(bmsg);
}


void closedevice_maybe_ioctl(cdev dev, bool restore_termio) {
  int fd = dev->fd;
  if(fd== -1) serror("Com device not open",1);
  vmsg("Closing device");
  if(restore_termio) {
    if (tcsetattr(fd, TCSADRAIN, &dev->orig_termio) < 0) {
      sprintf(msg,"Can't ioctl set device %s.\n",dev->filename->data);
      serror(msg,1);
    }
  }
  if(fd!=0)
    close(fd);
  dev->fd = -1;
}
void closedevice(cdev dev) {
  closedevice_maybe_ioctl(dev, true);
}

// Result close gfd ->  1: ok, but unlink of pty symlink failed
// Result close gfd ->  0: all fine
// Result close gfd -> -1: was not registered
// Result close gfd -> 0-(10+errno): system error

void doclose(void) {
  gettoken(LOWERCASE);
  if(b_seqc_Static(token,"hardcom")) {
    // TODO: "hardcom" reads like "emergency", so I'd expect ioctl rather at "com"
    closedevice_maybe_ioctl(&maindev, true);
  }
  else if(b_seqc_Static(token,"com")) {
    closedevice_maybe_ioctl(&maindev, false);
  }
  else if(b_seqc_Static(token,"file")) {
    if(filep==NULL) serror("Log file not open",4);
    fclose(filep);
    filep=NULL;
  }
  else if(b_seqc_Static(token,"gfd") || b_seqc_Static(token,"pty")) {
    int ptyfd=getfd(script, scriptspace, &pc);
    if (ptyfd < 0)
      serror("usage: \"close gfd !6\" for the seventh generic fd",5);
    resultcode = gfd_close(ptyfd);
  }
  else serror("close only takes hardcom, com, file or gfd(pty) argument",5);
}


// Historic behaviour:
// 
// opengt, when first invoked, chooses
//    * filename from CLI or if no filename given
//    * filename from env COMGTDEVICE or if variable not set
//    * filename from internal list
// open com, when first or later invoked, chooses
//    * filename from script line/string ("-" counts as stdin, this was wrong in manpage)
// opengt, when later invoked, chooses
//    * filename from last opened and closed device

// opengt actually sets COMGT_DEVICE while open com does not.

// Simulating all this historic behaviour here but I would rather deprecate it.

void opendevice(cdev cdev, bstring filename) {
  bool opengt_mode=(filename==NULL);
  
  if(cdev->fd >= 0) {
    serror("Device already opened", 7);
  }
  
  int mode; // 1 = no open(), 2 = open() the list, 3 = open() filename
  
  // PART I - filename
  if(opengt_mode) {
    if(cdev->filename) {
      filename = cdev->filename;
      mode=3; // CLI or last opened
    }
    else {
      char* devenv = getenv("COMGTDEVICE");
      if(devenv != NULL && strlen(devenv)) {
        filename=b_bfromc(devenv);
        mode=3; // env
      }
      else {
        mode=2; // no device on command line or env so try the list of devices
      }
    }
  }
  
  if(!opengt_mode) {
    if(   b_blength(filename)==0
       || b_beqc_Static(filename,"-")) {
      cdev->fd=0; // TODO collisions if multiple users? Warn/Fail?
      if(b_blength(filename)==0)
        filename=b_bfromc_Static("-");
      mode=1;
    }
    else {
      mode=3;
    }
  }
  
  // PART II - open the device/file
  if(mode==1) {
    // Do nothing, stdin already open
  }
  else if(mode==2) {
    printf("Trying list of devices\n");
    int dcount = 0;
    int newfd = -1;
    while(GTdevice[dcount]) {
      newfd = open(GTdevice[dcount], O_RDWR|O_EXCL|O_NONBLOCK|O_NOCTTY);
      if (newfd >= 0) break;
      dcount++;
    }
    if (newfd >= 0) {
      cdev->fd = newfd;
      b_bdestroy(cdev->filename);
      cdev->filename=b_bfromc_Ref(GTdevice[dcount]);
    }
    else {
      printf("Unable to locate default devices, try the -d option.\n");
      ext(1);
    }
  }
  else if(mode==3) {
    if ((cdev->fd = open(filename->data, O_RDWR|O_EXCL|O_NONBLOCK|O_NOCTTY)) <0) { //O_NONBLOCK|O_NOCTTY)) <0) {//
      sprintf(msg,"Can't open device %s.\n",filename->data);
      puts(msg);
      ext(1);
    }
    cdev->filename=filename;
  }
  
  
  // PART III - ioctl
  
  if (tcgetattr(cdev->fd, &cdev->orig_termio) < 0) {
    sprintf(msg,"Can't control (ioctl-get) device %s.\n",cdev->filename->data);
    serror(msg,1);
  }
  if (opengt_mode) {
    setenv("COMGTDEVICE",cdev->filename->data,1);
  }
  // our pre-/CLI-/... configured settings:
  int speed=get_speed(cdev);
  int our_flags = cdev->work_termio.c_cflag & (CSIZE | CSTOPB | CLOCAL | PARENB);
  
  cdev->work_termio = cdev->orig_termio;
  
  // replay our settings
  set_speed(cdev, speed);
  cdev->work_termio.c_cflag &= ~(CSIZE | CSTOPB | CLOCAL | PARENB);
  cdev->work_termio.c_cflag |=   our_flags;
  
  // our non-configurable settings
  cdev->work_termio.c_cflag |=  CREAD;
  cdev->work_termio.c_iflag &= ~(IGNCR | ICRNL | IUCLC | INPCK | IXON | IXANY | IGNPAR );
  cdev->work_termio.c_oflag &= ~(OPOST | OLCUC | OCRNL | ONLCR | ONLRET);
  cdev->work_termio.c_lflag &= ~(ICANON | XCASE | ECHO | ECHOE | ECHONL);
  cdev->work_termio.c_cc[VMIN] = 1;
  cdev->work_termio.c_cc[VTIME] = 0;
  cdev->work_termio.c_cc[VEOF] = 1;
  
  // commit settings
  setcom(cdev); 
  dormir(200000); /* Wait a bit (DTR raise) */
  sprintf(msg,"Opened %s as FD %d",cdev->filename->data,cdev->fd);
  vmsg(msg);
}



void doopen(void) {
  gettoken(LOWERCASE);
  int fdrequest = getfd(script, scriptspace, &pc);
  if(b_seqk(token,"com", 3)) {
    skipspaces();
    bstring devname;
    if(script[pc]=='$' || script[pc]=='\'' || script[pc]=='"') {
      devname = getstring();
    }
    else devname = gethardstring();
    if (b_blength(&token) == 3) { // "com"
      opendevice(&maindev, devname);
    }
    b_bdestroy(devname);
  }
  else if (b_seqc_Static(token,"file")) {
    bstring fname = getstring();
    if(fdrequest>=0) // GFD
      resultcode=openfile(fdrequest, fname->data);
    else {
      if(filep!=NULL) serror("Some main file is already open",4);
      if((filep=fopen(fname->data,"a"))==NULL) serror("Could not open file",4);
	}
    b_bdestroy(fname);
  }
  else if (b_seqc_Static(token, "pty")) {
    if (fdrequest < 0) serror("You didn't tell !<fdnum> for pty target",5);
    resultcode=openpty(fdrequest);
  }
  else if (b_seqc_Static(token, "recyclefd")) {
    if (fdrequest < 0) serror("You didn't tell !<fdnum> for recyclefd target",5);
    resultcode=recyclefd(fdrequest,getvalue());
  }
  else if (b_seqc_Static(token, "copygfd")) {
    if (fdrequest < 0) serror("You didn't tell !<fdnum> for copygfd target",5);
    int src = getfd(script, scriptspace, &pc);
    resultcode=gfd_copy(src, fdrequest);
  }
  else if (b_seqc_Static(token, "socketpair")) {
    if (fdrequest < 0) serror("You didn't tell !<fdnum> for socketpair target",5);
    int fdrequest2 = getfd(script, scriptspace, &pc);
    if (fdrequest < 0) serror("You didn't tell second !<fdnum> for socketpair target",5);
    resultcode=gfd_opensocketpair(fdrequest, fdrequest2);
  }
  else serror("OPEN only takes com, file, recyclefd, copygfd, pty or socketpair argument",5);
}

int doscript(void) {
  int a,b;
  int exitcode=0;
  pc=0;
  while(script[pc]) {
    if(script[pc]=='\n') {
      pc++;
      continue;
    }
    lastpc=pc;
    skipspaces();
    if(script[pc]=='\n') continue;
    if(verbose) printwhere();
    if(gettoken(LOWERCASE)) serror("Could not gettoken()",5);
    if(b_seqc_Static(token,"rem")) {
      skipline();
    }
    else if (b_seqc_Static(token,"label")) {
      skiptoken(); /* Get rid of keyword */
    }
    else if(b_seqc_Static(token,"setgfdname")) {
      int ptyfd=getfd(script, scriptspace, &pc);
      bstring gfdname = getstring();
      if (b_blength(gfdname) == 0)
        resultcode = gfd_clrname(ptyfd, NULL);
      else
        resultcode = gfd_setname(ptyfd, gfdname->data);
      b_bdestroy(gfdname);
    }
    else if(b_seqc_Static(token,"open")) {
      doopen();
    }
    else if(b_seqc_Static(token,"opengt")) {
      // Like >open com "filename"< but auto-select dev filename (despite other subtle differences)
      opendevice(&maindev, NULL); 
    }
    else if(b_seqc_Static(token,"close")) {
      doclose();
    }
    else if(b_seqc_Static(token,"exec")) {
      bstring cmd = getstring();
      strcpy(msg,"exec ");
      strcat(msg,cmd->data); /* Let sh do all the command line work! */
      b_bdestroy(cmd);
      execl("/bin/sh","sh","-c",msg,(char *)0);
      serror("Could not execl /bin/sh!",8);
    }
    else if(b_seqc_Static(token,"exit")) {
      lastpc=pc-1;
      ext(getvalue());
    }
    else if(b_seqc_Static(token,"testkey")) {
      dotestkey();
    }
    else if(b_seqc_Static(token,"kill")) {
      a=getvalue();
      resultcode=kill(getvalue(),a);
    }
    else if(b_seqc_Static(token,"fork")) {
      resultcode=fork();
    }
    else if(b_seqc_Static(token,"hset")) {
      hset=0;
      hstart=time(0);
      hset=htime()-getvalue();
    }
    else if(b_seqc_Static(token,"cd")) {
      bstring dir = getstring();
      resultcode=chdir(dir->data);
      b_bdestroy(dir);
    }
    else if(b_seqc_Static(token,"putenv")) {
      bstring arg = getstring();
      size_t l=b_blength(arg);
	  off_t i;
      for(i=0; i<l; i++) {
        char ch=arg->data[i];
        if(ch==0) serror("putenv command: name must not contain \\0!", 7);
        if(ch=='=') break;
      }
      if (i == l)
        resultcode=unsetenv(arg->data);
      else {
        arg->data[i++]=0;
        resultcode=setenv(arg->data, &(arg->data[i]), true);
      }
      b_bdestroy(arg);
    }
    else if(b_seqc_Static(token,"wait")) {
      resultcode=wait(0);
    }
    else if(b_seqc_Static(token,"system")) {
      bstring cmd = getstring();
      system(cmd->data);
      b_bdestroy(cmd);
    }
    else if(b_seqc_Static(token,"input")) {
      int infd = 0; // stdin
      int ptyfd=getfd(script, scriptspace, &pc);
      if (ptyfd >= 0)
        infd = gfds[ptyfd].fd;
      int index = getstringindex();
      bstring in = NULL;
      int res = doget_loop(&in, infd, "\n", 0, true);
      if (res >= 0) {
        if (stringvars[index]) b_bdestroy(stringvars[index]);
        stringvars[index] = in;
      }
      else b_bdestroy(in);
      resultcode = res; // can be 0 or <0
    }
    else if(b_seqc_Static(token,"get")) {
      doget();
    }
    else if(b_seqc_Static(token,"print")) {
      doprint(1);
    }
    else if(b_seqc_Static(token,"eprint")) {
      doprint(2);
    }
    else if(b_seqc_Static(token,"lprint")) {
      doprint(3);
    }
    else if(b_seqc_Static(token,"fprint")) {
      doprint(4);
    }
    else if(b_seqc_Static(token,"if")) {
      skipspaces();
      ifres=0;
      if(argisnumber()) {
        comparenumbers();
      }
      else {
        comparestrings();
      }
      if(!ifres) skipline();
    }
    else if(b_seqc_Static(token,"else")) {
      if(ifres) skipline();
    }
    else if(b_seqc_Static(token,"gosub")) {
      dogosub();
    }
    else if(b_seqc_Static(token,"return")) {
      if(preturn==0) serror("RETURN without gosub",5);
      pc=returns[--preturn];
    }
    else if(b_seqc_Static(token,"goto")) {
      dogoto();
    }
    else if(b_seqc_Static(token,"waitfor")) {
      resultcode=dowaitfor();
    }
    else if(b_seqc_Static(token,"waitquiet")) {
      resultcode=dowaitquiet();
    }
    else if(b_seqc_Static(token,"set")) {
      doset();
    }
    else if(b_seqc_Static(token,"dec")) {
      intvars[getintindex()]--;
    }
    else if(b_seqc_Static(token,"inc")) {
      intvars[getintindex()]++;
    }
    else if(b_seqc_Static(token,"let")) {
      dolet();
    }
    else if(b_seqc_Static(token,"dump")) {
      dodump();
    }
    else if(b_seqc_Static(token,"abort")) {
      vmsg("Aborting");
      abort();
    }
    else if(b_seqc_Static(token,"send")) {
      int ptyfd=getfd(script, scriptspace, &pc);
      bstring data = getstring();
      if (ptyfd >= 0)
        resultcode = gfd_write(ptyfd, data);
      else
        writecom(&maindev, data);
      b_bdestroy(data);
    }
    else if(b_seqc_Static(token,"flash")) {
      cdev cdev  = &maindev;
      int speed_restore=get_speed(cdev);
      set_speed(cdev,0);
      setcom(cdev);
      a=getdvalue();
      dormir(10000L*a);
      set_speed(cdev, speed_restore);
      setcom(cdev);
    }
    else if(b_seqc_Static(token,"sleep")) {
      a=getdvalue();
      if(a<10000) dormir(10000L*a);
      else sleep(a/100); /* I guess it's the same.  Oh well, past 100 secs,
                            use sleep instead.  */
    }
    else {
      /* Humour is the spice of life. */
      switch(time(0)&7) {
        case 0 : serror("That's human mumbo-jumbo to me",5); break;
        case 1 : serror("Lovely but incomprehensible",5); break;
        case 2 : serror("What's that, governor?",5); break;
        case 3 : serror("Very funny.  I don't get it",5); break;
        case 4 : serror("Huh?",5); break;
        case 5 : serror("comgt doesn't speak spanish",5); break;
        case 6 : serror("Mais, qu'est-ce que vous dites?",5); break;
        default: serror("%E-6837-% : Corrupted human data detected",5); break;
      }
    }
    skipspaces();
  }
  return(exitcode);
}

int main(int argc,char **argv) {
  unsigned int a;
  int aa,b,i,skip_default;
  unsigned char ch;
  unsigned char terminator='\n';
  char line[STRINGL];


  FILE *fp;
  hstart=time(0);
  hset=htime();
  preturn=0;
  skip_default=0;
  filep=NULL;
  init_vars();
  init_gfd();
  scriptspace=4096;
  tcgetattr(1, &cons);
  if((script=( char *)malloc(scriptspace))==NULL) {
    serror("Could not malloc()",3);
  }
  script[0]=0;
  pc=0;
  lastpc=0;
  for(a=0;a<NVARS;a++) {
    intvars[a]=0;
    stringvars[a] = NULL;
  }
  
  init_comgt_dev_struct(&maindev);
  scriptfile[0]=0;
  b=0; a=0;
  for(a=0;a<strlen(argv[0]);a++) {
    if(argv[0][a]=='/') b=a+1;
  }
  while((aa=getopt(argc,argv,"xheVvd:t:sb:"))!= -1) {
    switch(aa) {
      case 0:
        ext(0);
        break;
      case 't':
        terminator=optarg[0]; // TODO: accept escape expression
        sprintf(msg,"Alternate line terminator set to \"%c\"",terminator);
        vmsg(msg);
        break;
      case 'd':
        maindev.filename=b_bfromc_Ref(optarg);
        break;
      case 'e':
        comecho=1;
        vmsg("Communication echo turned on");
        break;
      case 'v':
        verbose=1;
        vmsg("Verbose output enabled");
        break;
      case 's':
        skip_default=1;
        break;
      case 'V':
        printf("comgt version %s Copyright %s (c) %s\n", comgt_version,AUTHORS,YEARS);
        ext(1);
        break;
      case 'h':
        printf("comgt version %s Copyright %s (c) %s\n", comgt_version,AUTHORS,YEARS);
        printf("\nType 'comgt help' for more information\n");
        ext(1);
        break;
      case 'x':
        printf("High speed overide (115200 is now 57600).\n");
        maindev.high_speed=true;
        break;
      default:
        ext(1);
    }
  }
  if(optind<argc) {
    strcpy(scriptfile,argv[optind++]);
    sprintf(msg,"Script file: %s",scriptfile);
    vmsg(msg);
  }

  char * code;
  code = get_code(scriptfile);
  if (code != NULL){
    scriptspace=strlen(code)+2;
    if((script=( char *)realloc(script,scriptspace))==0) {
       serror("Could not allocate memory for script.",3);
       }
       strcpy(script,code);
      for(aa=0;aa<scriptspace;aa++) {
         if(script[aa]==terminator) script[aa]='\n';
      }
      //scriptfile[0] = '\0';
    }
  else if (scriptfile[0]) {
  	code = get_code("default");
  	if (code != NULL && !skip_default){
    		scriptspace=strlen(code)+2;
    		if((script=( char *)realloc(script,scriptspace))==0) {
       			serror("Could not malloc()",3);
       		}
       		strcpy(script,code);
       		for(aa=0;aa<scriptspace;aa++) {
         		if(script[aa]==terminator) script[aa]='\n';
      		}
      //scriptfile[0] = '\0';
    }
    if((fp=fopen(scriptfile,"r"))==NULL) {
		strcpy(scriptfilepath,"/etc/comgt/");
		strcat(scriptfilepath,scriptfile);
		if((fp=fopen(scriptfilepath,"r"))==NULL) {
        sprintf(msg,"Could not open scriptfile \"%s\"",scriptfile);
      		serror(msg,1);
		}
    }
    i=strlen(script);
    if(i) {
      script[i++]='\n'; script[i]=0;
    }
    while((fgets(line,STRINGL-1,fp))!=NULL) {
      b=strlen(line);
      if((scriptspace-i)<STRINGL) {
        scriptspace+=STRINGL+STRINGL;
        if((script=(char *)realloc(script,scriptspace))==NULL) {
          serror("Could not realloc()",3);
        }
      }
      for(aa=0;aa<b;aa++) {
        script[i]=line[aa];
        if(script[i]==terminator) script[i]='\n';
        i++;
      }
    }
    script[i]=0;
    fclose(fp);
  }

  if(script[0]) {
    // Why remove last token/expression? Would better add newline.
    i=strlen(script)-1;
    while((script[i]=='\n' || script[i]==' ' || script[i]=='\t') && i!=0) i--;
    script[++i]='\n';
    script[++i]=0;
  }
  scriptspace=strlen(script); /* Script is one huge string */
  // I wouldn't call me a bean counter, but this is the right place to free some space!
  if((script=(char *)realloc(script,scriptspace+1))==NULL) {
    // No problem on fail, no serror()
    vmsg("Could not realloc() (?? this error is really unprobable)");
  }
  
  /* Indexing labels */
  label=(struct tagbstring*)malloc(sizeof(label  [0])*MAXLABELS);
  if(label==NULL) {
    serror("Can't malloc",3);
  }
  labels=0;
  pc=0;
  while(script[pc]) {
    lastpc=pc;
    gettoken(LOWERCASE);
    if(b_seqc_Static(token,"label")) {
      gettoken(ALLCASE);
      for(aa=0;aa<labels;aa++) {
        if(b_seqs(token,label[aa])) {
          pc=lastpc;
          serror("Duplicate label",5);
        }
      }
      if(labels==MAXLABELS) serror("Maximum number of labels reached",3);
      label[labels]=token; // Copies the reference to script[lastpc+?]+len
      labels++;
    }
    skipline();
  }
  //printf(script);
  if(verbose) {
    sprintf(msg,"argc:%d",argc);
    vmsg(msg);
    for(aa=0;aa<argc;aa++) {
      sprintf(msg,"argv[%d]=%s",aa,argv[aa]);
      vmsg(msg);
    }
    vmsg(" ---Script---");
    aa=0; b=0; ch='\n';
    while(aa<scriptspace) {
      if(ch=='\n' && script[aa]!=0) {
        fprintf(stderr,"%4d@%04d ",++b,aa);
      }
      ch=script[aa++];
      fputc(ch,stderr);
    }
    vmsg(" ---End of script---");
  }

  if(script[0]==0) {
    fprintf(stderr,"No script!\n");
    ext(1);
  }

  a=doscript();
  dormir(200000);
  if(maindev.fd!= -1) close(maindev.fd);
  sprintf(msg,"Exit with code %d.\n",a);
  vmsg(msg);
  ext(a);
}

#endif // #ifndef NOCODE

#endif // #ifndef COMGT_MAIN
