#!/bin/sh

if [ \! -e webrelease.cfg ]; then
	printf >&2 "%s: Error: There must be a ./webrelease.cfg in cwd!\n" "$0"
fi

ret() {
	return $1
}
nr=-
checksuccess() {
	[ $? -eq 0 ] && return;
	fail 77 "[line $nr] $1"
}
fail () {
	perror "$2"
	ret $1; exit
}
perror() {
	printf >&2 "%s: Error: %s\n" "$0" "$1"
}


# version number
version="`grep -E '^#define COMGT_VERSION ' c/version.h`"
version="${version###define COMGT_VERSION }"
printf "%s: Releasing version %s\n" "$0" "$version"
tmp="`git tag -l "comgt-$version"`"
[ -n "$tmp" ] && fail 50 "The version was already released, choose another version number!"



codepath="$PWD"
releasepath=
buildpath=
do_dstpath() {
	if [ -n "$releasepath" ]; then
		printf >&2 "%s: Error at line %d: You cannot set dstpath twice!\n" "$0" $nr
		ret 51; exit
	fi
	
	releasepath="$dstpath"
}

do_cmdline() {
	printf "%s: cmdline: %s\n" "$0" "$cmdline"
	
	cmd="${cmdline%% *}"
	cmd="${cmd%%	*}" # tabulator
	
	args="`printf %s "$cmdline" | sed -r "s#^[ \t]*$cmd[ \t]+##"`"
	
	if [ \! _"${cmd#!}" = _"$cmd" ]; then
		cmdline="${cmdline#*!}"	
		eval "$cmdline"
		return
	fi
	
	if [ _"$cmd" = _setenv ]; then
		export $args
		checksuccess "Error exporting."
		return
	fi
	
	if [ _"$cmd" = _prepdir ]; then
		mkdir -p "$releasepath"; checksuccess "Destination path not present and uncreatable"
		cd -- "$releasepath"; checksuccess "(Rare error) Destination path not accessible"
		mkdir -p tmpbuild
		buildpath=tmpbuild/"comgt-$version"
		if [ -e "$buildpath" ]; then
			rm -rf "$buildpath"; checksuccess "Couldn't remove $PWD//$buildpath"
		fi
		mkdir "$buildpath"
		cd "$buildpath"
		cp -a "$codepath/.git" .
		git reset --hard
		return
	fi
	
	[ -n "$buildpath" ]; checksuccess "\"C: prepdir\" must be first command"
	
	if [ _"$cmd" = _savedir -o _"$cmd" = _savebin ]; then
		filename="`printf "$args" "$version"`"
	fi
	if [ _"$cmd" = _savedir ]; then
		(cd ..; tar -czf ../"$filename" comgt-"$version")
		checksuccess "Error when packing directory."
		return
	fi
	if [ _"$cmd" = _savebin ]; then
		cp comgt ../../"$filename"
		checksuccess "Error when copying binary."
		return
	fi
	
	if [ _"$cmd" = _configure ]; then
		bstring_src="`readlink -f "$codepath/gen/bstring-source"`"
		checksucces "You must ./configure.sh in the main source"
		bstring_test="${bstring_src#../}"
		bstring_test="${bstring_test#./}"
		[ \! _"$bstring_test" = _"$bstring_src" ] && bstring_src="$codepath/gen/$bstring_src"
		./configure.sh --bstring-src="$bstring_src"
		checksuccess "Error when configuring."
		return
	fi
	if [ _"$cmd" = _makebin ]; then
		make clean all
		checksuccess "Error when building."
		return
	fi
	
	printf >&2 "%s: Error at line %d: Don't know command '%s'!\n" "$0" $nr "$cmd"
	ret 52; exit
}

nr=0
while read line; do
	nr=$((nr+1))
	line="${line## }"
	line="${line##	}" # (tabulator)
	
	cmdline="${line#C: }"
	if [ \! :"$cmdline" = :"$line" ]; then
		do_cmdline && continue
	fi
	
	dstpath="${line#dstpath=}"
	if [ \! :"$dstpath" = :"$line" ]; then
		do_dstpath && continue
	fi
	
	line="${line%%#*}"
	[ -z "$line" ] && continue
	
	printf >&2 "%s: Error: cann't recognize line '%s'!\n" "$0" "$line"
	
	ret 53
	exit
done <webrelease.cfg


cd "$codepath"
git tag comgt-"$version"; checksuccess "Release OK, but couldn't tag version!"
rm -rf "$releasepath/$buildpath"

printf "Release done."

